"""
A set of simple worlds that can be used to evaluate Becca agents

These are toy worlds, but they are diverse.
Their purpose is to push the limits of typical learning agents
is as simple a manner as possible.
"""
